##############################################
## matCrossprod ##############################
## Function to generate cross-product of X & Z
##############################################

matCrossprod = function(X,Z){
  out = matrix(0,dim(X)[1],dim(X)[2]*dim(Z)[2])
  for(i in 1:dim(X)[2]){
    x = X[,i]
    for(j in 1:dim(Z)[2]){
      z = Z[,j]
      out[,dim(Z)[2]*(i-1)+j] = x*z  
    }
  }
  return(out)
}


################################################################
## chisq.max.full ##
## Function to compute maxmimum of LR across threshold parameter
################################################################

chisq.max.full = function(p,y,X,oo,
                n.initials,ind.initials){

    X1 = cbind(1,X)
    glm.null = glm(y ~ X, family=binomial(link='logit'))
    loglike.null = glm.null$dev # equals to -log(Lhat) 
    LRcandid = sapply(1:n.initials,function(s){
                    Wj = X1[,oo] * as.numeric(ind.initials[,s])
                    out = glm.alt(y=y,XX=cbind(X,Wj),
                                loglike.null=loglike.null)
                    return(out)
                })
    LRcandid = LRcandid[!is.na(LRcandid)]
    LR = max(LRcandid)
    return(LR)
}


#######################################
## glmnet.twostage ####################
### glm model with feature selection ##
#######################################

glmnet.twostage.full = function(M,y,p,q,s,M.penalty,interaction=c('yes','no')){
   
    res = glmnet(M,y,family="binomial",alpha=0.5,penalty.factor=M.penalty) 
    if(interaction=='no'){
        check = abs(res$beta[p+(1:q),])
      
        
    }else{
        #check = abs(res$beta[p+q+(1:(p*q)),])
        if(s==(1+p)) check = abs(res$beta[(p+q+1):(ncol(M)),]) else {
         check = abs(res$beta[(p+1):(ncol(M)),])
        }
    }
    check = colSums(check)
    check = sum(check==0)
    l.dr = length(res$dev.ratio)
    dr.gap = abs(res$dev.ratio[(check+1):(l.dr-1)]-res$dev.ratio[(check+2):l.dr])/res$dev.ratio[(check+1):(l.dr-1)]
    lambda = res$lambda[check+which.max(dr.gap)]
    res = glmnet(M,y,family="binomial",alpha=0.5,lambda=lambda,penalty.factor=M.penalty)
    return(list(a0=res$a0,beta=res$beta))
}


##################
## glm.alt ##
## compute LR ####
glm.alt = function(yy,XX,loglike.null){
    fit = keepWarnings(glm(yy ~ XX, family=binomial(link='logit')))
    if(length(fit$warning) == 0){
        loglike.alt = fit$value$dev
        out = loglike.null - loglike.alt
    }else{
        out = NA
    }
    return(out)
}
#######################################################################
### compute.teststat.full #############################################
### Function to compute LR statistics for a two-step threshold modeling

## Model: logit P(y=1|X,Z)=beta1^T(1,X)+beta2^TXs I(gamma^T(1,Z)>0), and we
## test for H_0: beta2=0


## Input:

## y: binary disease outcome (taking value 0,1)
## X: covariates for modeling disease risk model
## oo: indicator of a subset of (1,X): Xs=(1,X)[oo] where it is of interest to 
##      test whether the relationship between y and Xs differ according to subgroups 
##     defined by Z
## Z: covariates (biomarkers) for defining subgroup with differential risk models
## interaction: whether univariate marker combination is based on main effect 
##         for Z or interaction between Xs and Z
## n.initials, number of thresholds for the univariate combination score
## lb.quantile, ub.quantile, quantiles for selecting threshold parameters
## sel: indicator of whether feature selection would be implemented

## Output: 
## LR: maximum of Weighted LR (WLR) test statistics 
######################################################################################


compute.teststat.full = function(X,Z,y,oo=NULL,interaction=c('yes','no'),n.initials,lb.quantile=.1, ub.quantile=.9,sel=T){

    s=length(oo) 
    p=ncol(as.matrix(X))  ## dimension of X 
    q=ncol(as.matrix(Z))  ## dimension of Z
    if(interaction=='no'){  ## estimate gamma^TZ based on main effect of Z only
      if (sel==T){
        # perform feature selection in estimating gamma^TZ if either sel=T or p>1 
        beta.XZ = glmnet.twostage.full(M=cbind(X,Z),y,p=p,q=q,s=s,M.penalty=c(rep(0,p),rep(1,q)),
                interaction='no')
        Z.beta = Z %*% beta.XZ$beta[-c(1:p)]
      } else {
        # no feature selection when estimating gamma^T
        beta.XZ=glm.fit(cbind(1,X,Z),y,family=binomial())$coef[-1]
        Z.beta = Z %*% beta.XZ[-c(1:p)]
      }
    }else{
      ## estimate gamma^TZ based on interaction between X and Z 
      XZcrossprod = matCrossprod(as.matrix(cbind(1,X)[,oo]), Z)  
       if (sel==T | s>2){
       # perform feature selection in estimating gamma^TZ if either sel=T or s>2 
       beta.XZXZ = glmnet.twostage.full(M=cbind(X,XZcrossprod),y,p=p,q=q,s=s,M.penalty=c(rep(0,p),rep(1,s*q)),
                interaction='yes')  
        Z.beta = cbind(XZcrossprod) %*% beta.XZXZ$beta[-c(1:p)]
      } else {
        # no feature selection when estimating gamma^TZ
         beta.XZXZ = glm.fit(cbind(1,X,XZcrossprod),y,family=binomial())$coef[-1]
        Z.beta=Z%*%beta.XZXZ[-c(1:(p+(s-1)*q))] 
      }
    }
    tau.candid = quantile(Z.beta,seq(lb.quantile,ub.quantile,length=n.initials))
    ind.initials = sapply(tau.candid,function(x){return(Z.beta > x)})
 
    ## compute max of LR over various thresholds for estimated gamma^TZ
    teststat = chisq.max.full(p,y,X,oo,n.initials,ind.initials)  
    
    return(teststat)
}
